package asd.graphe;

import asd.base.PileStatique;

/**
 * Une classe pour les composantes fortement connexes d'un graphe orient.
 * (voir le <a href="../../../html/CFC.html">source</a>).
 * @author  2002 Marc Gatano
 * @version 3.0 (26/09/02)
 */
public class CFC {

    private GrapheOriente G;  // le graphe orient
    private GrapheOriente TG; // TG = transpos(G)
    
    private int[] c;   // la couleur des sommets

    private static final int blanc = 0;
    private static final int gris  = 1;
    // la couleur noir est inutile ici

    /**
     * Cre un objet CFC pour les composantes fortement connexes du
     * graphe G.
     */
     public CFC(GrapheOriente G) {
        this.G = G;
	this.TG = new GrapheOriente(G.ordre());
        this.c = new int[G.ordre()];
    }

    /**
     * Retourne un tableau cfc tel que cfc[i] = k si et seulement si k est 
     * le numro de la composante connexe de i, pour tout i dans [0..n-1] et 
     * pour k dans [1..n], o n est l'ordre du graphe G.
     * Complexit : O(S + A) o S et A sont respectivement le nombre de
     * sommets et d'arcs du graphe G.
     */
    public int[] cfc() {
	return cfc(topo());
    }

    /**
     * Ralise le parcours en profondeur du graphe TG, en examinant les
     * sommets dans l'ordre o ils sont dpils de la pile (l'ordre
     * calcul sur le graphe G) et numrote les arbres de la fort en
     * profondeur correspondante.
     * Complexit : O(S + A) o S et A sont respectivement le nombre de
     * sommets et d'arcs du graphe TG (et du graphe G).
     */
    private int[] cfc(PileStatique pile) {
	int n = 1;                      // le numro de la CFC courante
        int[] cfc = new int[G.ordre()]; // cfc[i] = k <==> le sommet i est
                                        // dans la CFC numro k
        for ( int i = 0; i < TG.ordre(); i++ ) {
            c[i] = blanc;
        }
	// les sommets de TG sont visits dans l'ordre topologique sur G.
	//
	while ( nonVide(pile) ) {
	    int i = depiler(pile);
            if ( c[i] == blanc ) {  // i n'a jamais t atteint ainsi que
                visiter(i,n++,cfc); // tous ses descendants,  qui on va
            }                       // attribuer la CFC numro n.
        }
        return cfc;
    }
    
    /**
     * Visite le sommet i et ses descendants dans le graphe TG, et attribue
     *  tous ces sommets la CFC numro n.
     * Prcondition : i est blanc.
     * Postrelation : i n'est pas blanc et le numro de CFC n a t attribu
     *  tous les sommets accessibles depuis i (i inclus).
     * Complexit : O(Ai) o Ai est le nombre d'arcs accessibles depuis i.
     */
    private void visiter(int i, int n, int[] cfc) {
        c[i] = gris;   // i vient d'tre dcouvert
	cfc[i] = n;    // i appartient  la CFC courante (numro n)
        Adjacents adj = TG.adjacents(i);
        
        while ( adj.nonVide() ) {
            int s = adj.suivant();
            if ( c[s] == blanc ) {
                visiter(s,n,cfc);
            }
        }
        // ( c[i] = noir ) fin de tratement de i
    }

    /**
     * Trie le graphe G suivant l'ordre des dates de fin dcroissantes et
     * retourne une pile contenant tous les sommets du graphe, telle que
     * pour tout sommet u et v dans la pile, si u est dpil avant v alors
     * soit u et v ne sont pas connects, soit il existe dans G un chemin
     * de u  v. La mthode topo initialise galement le graphe transpos
     * TG.
     * Complexit : O(S + A) o S et A sont respectivement le nombre de
     * sommets et d'arcs du graphe G (et du graphe TG).
     */
    private PileStatique topo() {
	PileStatique pile = new PileStatique();
        
        for ( int i = 0; i < G.ordre(); i++ ) {
            c[i] = blanc;
        }

        for ( int i = 0; i < G.ordre(); i++ ) {
            if ( c[i] == blanc ) {
                visiter(i,pile);
            }
        }
	return pile;
    }

    /**
     * Visite le sous-graphe de G de sommet i en profondeur et empile tous les
     * sommets du sous-graphe de racine i dans la pile, dans l'ordre des dates
     * de fin dcroissantes. Connecte le graphe transpos TG.
     * Prcondition : i est blanc.
     * Postrelation : i n'est pas blanc et pour tout j accessible depuis
     * i (i inclus), j a t empil avant i.
     * Complexit : O(Ai) o Ai est le nombre d'arcs accessibles depuis i.
     */
     private void visiter(int i, PileStatique pile) {
        c[i] = gris;
        Adjacents adj = G.adjacents(i);
        while ( adj.nonVide() ) {
            int s = adj.suivant();
	    TG.connecter(s,i);     // on construit TG
            if ( c[s] == blanc ) {
                visiter(s,pile);   // visite "en profondeur"
            }
        }
        // ( c[i] = noir ) fin de traitement de i
	empiler(i,pile);
    }

    // pour se simplifier la vie

    private void empiler(int i, PileStatique pile) {
        pile.empiler(new Integer(i));
    }

    private int depiler(PileStatique pile) {
        int i = ((Integer) pile.sommet()).intValue();
        pile.depiler();
        return i;
    }

    private boolean nonVide(PileStatique pile) {
        return ! pile.vide();
    }
}
