package asd.graphe;

import asd.base.ObjetAClef;
import asd.tri.TasCroissant;

/**
 * Une classe pour l'algorithme de Kruskal.
 * (voir le <a href="../../../html/Kruskal.html">source</a>).
 * @author  2005 Marc Gatano
 * @version 4.0 (18/04/05)
 */
public class Kruskal {

    private GrapheNonOrienteValue G; // l'ACM  construire

    // attributs du parcours en profondeur

    private int[] p; // p[j] = i <==> i est le pre de j
    private int[] c; // la couleur des sommets

    private static final int blanc = 0;
    private static final int gris  = 1;
    private static final int noir  = 2;

    /**
     * Construit un objet Kruskal pour calculer un arbre couvrant de
     * poids minimum du graphe non orient connexe valu G.
     * Complexit : O(1).
     */
    public Kruskal(GrapheNonOrienteValue G) {
	this.G = G;
    }

    /**
     * Retourne un arbre couvrant de poids minimum du graphe G.
     * Complexit : O(n.lg(n)) o n est le nombre d'artes de G.
     */
    public GrapheNonOriente acm() {

        GrapheNonOriente Gp = new GrapheNonOriente(G.ordre());
        TasCroissant T = new TasCroissant(G.aretes());
        initialiser(T);

	Composante CC = new Composante(G.ordre()); // un sommet par cc
        int k = G.ordre();                         // le nombre de cc

	while ( k > 1 ) {

            Arete uv = (Arete) T.extraireMinimum().info();

	    int u = uv.origine();
	    int v = uv.extremite();

	    if ( CC.separes(u,v) ) {
		Gp.connecter(u,v);
		CC.reunir(u,v); 
                k--;
	    }
	}
        // INVARIANT : 
        // - le graphe Gp est un sous-ACM de G.
        // - u et v sont dans la mme cc <==> u <--> v dans Gp
	// - k > 1 ==> il existe CC1 et CC2 deux CC et u1 dans CC1
        //   et u2 dans CC2 avec (u1,u2) arte de G dans T
	// VARIANT :
        // (k,|G|) dcrot et est minor par (1,0)
        //
	return Gp;
    }

    /**
     * Effectue un parcours en profondeur du graphe G et ajoute chaque
     * arte de G une et une seule fois dans le tas T..
     * Complexit : O(S + A) o S est l'ordre du graphe et A le nombre
     * d'artes.
     */
    private void initialiser(TasCroissant T) {
        c = new int[G.ordre()];
        p = new int[G.ordre()];
	for ( int i = 0; i < G.ordre(); i++ ) {
	    c[i] = blanc;
	    p[i] = -1;
	}
        visiter(0,T);
    }

    /**
     * Visite le sous-graphe de sommet i en profondeur et ajoute chaque arte
     * visite une et une seule fois dans le tas T.
     * Prcondition : i est blanc.
     * Postrelation : i est noir et chaque arte visite a t ajoute une et
     * une seule fois dans le tas T.
     * Complexit : O(Ai) o Ai est le nombre d'artes accessibles depuis i.
     */
    private void visiter(int i, TasCroissant T) {
	c[i] = gris;
	Adjacents adj = G.adjacents(i);
	
	while ( adj.nonVide() ) {
	    int s = adj.suivant();

	    if ( c[s] == blanc ) {
                T.ajouter(new ObjetAClef(new Arete(i,s),
                                         G.valuation(i,s)));
		p[s] = i;
		visiter(s,T);
	    }
            else {
                if ( c[s] == gris && s != p[i] ) {
                    T.ajouter(new ObjetAClef(new Arete(i,s),
                                             G.valuation(i,s)));
                }
            }
	}
	c[i] = noir;
    }
}
