package asd.graphe;

import asd.base.PileStatique;

/**
 * Une classe pour le tri topologique d'un graphe orient acyclique.
 * (voir le <a href="../../../html/TriTopologique.html">source</a>).
 * @author  2002 Marc Gatano
 * @version 2.0 (02/12/02)
 */
public class TriTopologique {

    private GrapheOriente G;
    
    private int[] c; // la couleur des sommets

    private static final int blanc = 0;
    private static final int gris  = 1;
    private static final int noir  = 2;

    /**
     * Cre un objet TriTopologique pour trier topologiquement le
     * graphe G.
     * Prcondition : G est acyclique.
     */
    public TriTopologique(GrapheOriente G) {
	this.G = G;
	this.c = new int[G.ordre()];
    }

    /**
     * Trie le graphe topologiquement et retourne une pile contenant tous les
     * sommets du graphe, telle que pour tout sommet u et v dans p, si u est
     * dpil avant v alors soit u et v ne sont pas connects, soit il existe
     * dans G un chemin de u  v.
     * Complexit : O(S + A) o S est l'ordre du graphe et A le nombre d'arcs.
     */
    public PileStatique trier() {
        PileStatique pile = new PileStatique(G.ordre());
	
	for ( int i = 0; i < G.ordre(); i++ ) {
	    c[i] = blanc;
	}
        // tous les sommets sont blancs et aucun sommet n'a de pre

	for ( int i = 0; i < G.ordre(); i++ ) {
	    if ( c[i] == blanc ) {
		visiter(i,pile);
	    }
	}
        return pile;
    }

    /**
     * Visite le sous-graphe de sommet i en profondeur et empile tous les
     * sommets du sous-graphe de racine i dans la pile, dans l'ordre inverse
     * de l'ordre topologique.
     * Prcondition : i est blanc.
     * Postrelation : i est noir et tous les sommets accessibles depuis
     * i (i inclus) ont t empil dans la pile, dans l'ordre inverse de
     * l'ordre topologique.
     * Complexit : O(Ai) o Ai est le nombre d'arcs accessibles depuis i.
     */
    private void visiter(int i, PileStatique pile) {
	c[i] = gris;
	Adjacents adj = G.adjacents(i);
	
	while ( adj.nonVide() ) {
	    int s = adj.suivant();

            // dbut de tratement
            if ( c[s] == blanc ) {
		visiter(s,pile);
            }
            // fin de tratement
	}
	c[i] = noir;
	// pour tout sommet v dans pile, f[u] > f[v]
        pile.empiler(new Integer(i));
    }
}
